"""django-lrucache-backend - A smarter local memory cache backend for Django"""

from .backend import LRUObjectCache

__version__ = "3.0.1"
__author__ = "Josh Smeaton <josh.smeaton@gmail.com>"
__all__ = ["LRUObjectCache"]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import time
from threading import RLock

from django.core.cache.backends.base import DEFAULT_TIMEOUT, BaseCache
from lru import LRU  # dependency: lru-dict

"Thread-safe in-memory LRU object cache backend."
# Global in-memory store of cache data. Keyed by name, to provide
# multiple named local memory caches.
_caches = {}
_locks = {}


class LRUObjectCache(BaseCache):
    """
    A local memory cache that:

        1. Avoids serialization and deserialization
        2. Implements an LRU eviction strategy
        3. Implements timeouts on add()/get()/has_key()

    Uses a C based LRU dict for performance. See: https://github.com/amitdev/lru-dict

    Honours `MAX_ENTRIES` but not `CULL_FREQUENCY`, as the LRU algorithm will
    evict the least recently used value as required.
    """

    def __init__(self, name, params):
        BaseCache.__init__(self, params)
        self._cache = _caches.setdefault(name, LRU(self._max_entries))
        self._lock = _locks.setdefault(name, RLock())

    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        new_key = self.make_key(key, version=version)
        with self._lock:
            if self._has_expired(key, version=version):
                self._set(new_key, value, timeout)
                return True
            return False

    def get(self, key, default=None, version=None):
        value, timeout = self._get(key, default, version)
        return value

    def _get(self, key, default=None, version=None):
        key = self.make_key(key, version=version)
        with self._lock:
            value, expiration = self._cache.get(key, (default, -1))
        if not self._is_expired(expiration):
            return value, expiration
        with self._lock:
            try:
                del self._cache[key]
            except KeyError:
                pass
            return default, -1

    def _set(self, key, value, timeout=DEFAULT_TIMEOUT):
        timeout = self.get_backend_timeout(timeout)
        self._cache[key] = (value, timeout)

    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        with self._lock:
            self._set(key, value, timeout)

    def incr(self, key, delta=1, version=None):
        with self._lock:
            # remain locked the entire time so the incr is guaranteed to be correct
            value, exp = self._get(key, version=version)
            if value is None:
                raise ValueError("Key '%s' not found" % key)
            new_value = value + delta
            new_key = self.make_key(key, version=version)
            self._cache[new_key] = (new_value, exp)
        return new_value

    def has_key(self, key, version=None):
        # _has_expired will do the locking
        if not self._has_expired(key, version=version):
            return True

        with self._lock:
            try:
                del self._cache[key]
            except KeyError:
                pass
            return False

    def delete(self, key, version=None):
        key = self.make_key(key, version=version)
        with self._lock:
            try:
                del self._cache[key]
            except KeyError:
                pass

    def clear(self):
        with self._lock:
            self._cache.clear()

    def _has_expired(self, key, version=None):
        _, exp = self._get(key, version=version)
        return self._is_expired(exp)

    def _is_expired(self, expiration):
        if expiration is None or expiration > time.time():
            return False
        return True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lrucache_backend']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.2', 'lru-dict>=1.1,<2.0']

setup_kwargs = {
    'name': 'django-lrucache-backend',
    'version': '3.0.1',
    'description': 'A smarter local memory cache backend for Django',
    'long_description': "django-lrucache-backend\n=======================\n\n.. image:: https://img.shields.io/pypi/v/django-lrucache-backend.svg\n    :target: https://pypi.python.org/pypi/django-lrucache-backend\n    :alt: Latest PyPI version\n\n.. image:: https://travis-ci.org/kogan/django-lrucache-backend.svg?branch=master\n   :target: https://travis-ci.org/kogan/django-lrucache-backend\n   :alt: Latest Travis CI build status\n\nA smarter local memory cache backend for Django.\n\n.. image:: benchmarking/2.0.0/objects-get.png\n   :alt: Cache Performance\n\n`Set Performance <benchmarking/2.0.0/objects-set.png>`_. `Delete performance\n<benchmarking/2.0.0/objects-delete.png>`_.\n\nAbout\n-----\n\n`lrucache_backend` is an in-memory cache that improves upon the existing\n`LocMemCache` that Django provides.\n\nComes with cache timeouts and a smart eviction strategy that prefers to keep\nkeys that are used often and evict keys that are not.\n\nOriginally developed to avoid poorly reimplementing local object stores for\nservice layer objects. For example:\n\n.. code-block:: python\n\n    def get_data_before(self):\n        if not hasattr(self, '__data'):\n            self.__data = self.expensive_query()\n        return self.__data\n\n    def get_data_after(self):\n        lcache = caches['local']\n        data = lcache.get('our_data')\n        if not data:\n            data = self.expensive_query()\n            lcache.set('our_data', data, timeout=600)\n        return data\n\nThe benefits (despite the longer method) include timeouts, sharing data between\nrequests, and avoiding network requests. This is especially useful when there\nare hundreds or thousands of property accesses that would hit the cache where\nnetwork overhead would be prohibitive. The `Fat model` pattern can greatly\nbenefit from tiered caching.\n\nGood for?\n^^^^^^^^^\n\nAn in memory cache is good for small data that changes rarely. It's effectively\na global dictionary shared between requests in the same process. Small lookup\ntables and database backed settings are good candidates.\n\nA small number of keys should be used to avoid engaging the culling strategy\nof the cache. Performance goes down fast as soon as the maximum number of keys\nare reached, and keys start to evict.\n\nThis should **not** be used as your primary cache, but it makes for an\nexcellent secondary cache when you want to avoid the overhead of a network call.\n\nUse for:\n\n- Small lookup tables\n- Settings\n- Backing store for your service objects\n- Remembering values for the duration of a request or celery task\n- Small global template fragments like sidebars or footers\n- Secondary cache\n\nBad for?\n^^^^^^^^\n\nAn in memory cache is terrible for data that changes often. Because the cache\nis process local, it's extremely difficult to coordinate cache invalidation\nfrom external processes. For that reason, this library does nothing to support\ncache invalidation.\n\nThe cache shares memory with the application, so it's extremely important to\navoid storing a lot of keys, or any large values.\n\nDo **not** use for:\n\n- Instance attributes/properties\n- Full templates\n- Tables with a large number of rows\n- Large values\n- Large lists\n- Primary cache\n\nDifferences from LocMemCache\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n- Avoids pickling\n- Avoids key name validation\n- Uses an LRU eviction algorithm rather than a random percentage culling strategy\n\nInstallation\n------------\n\n.. code-block:: bash\n\n     pip install django-lrucache-backend\n\nRequirements\n^^^^^^^^^^^^\n\n* `lru-dict <https://pypi.python.org/pypi/lru-dict/>`_.\n\n`lru-dict` is implemented in C and is unlikely to work with non-CPython\nimplementations. There *are* compatible pure python libraries. If you need this\nability, please open an Issue!\n\nUsage\n-----\n\nConfigure your `CACHES` Django setting appropriately:\n\n.. code-block:: python\n\n    CACHES = {\n        'local': {\n            'BACKEND': 'lrucache_backend.LRUObjectCache',\n            'TIMEOUT': 600,\n            'OPTIONS': {\n                'MAX_ENTRIES': 100\n            },\n            'NAME': 'optional-name'\n        }\n    }\n\nAnd then use the cache as you would any other:\n\n.. code-block:: python\n\n    >>> from django.core.cache import caches\n\n    >>> local = caches['local']\n    >>> local.set('key', 123)\n    >>> local.get('key')\n    ... 123\n\nIf you're going to use this cache backend, then it's highly recommended to use\nit as a non-default cache. That is, do not configure this cache under the\n`default` name.\n\nLocal memory caches compete for memory with your application so it's in your\nbest interests to use it as sparingly and deliberately as possible.\n\nCompatibility\n-------------\n\nDjango 2.2+\nPython 3.6+\n\nLicence\n-------\n\nMIT\n\nAuthors\n-------\n\n`django-lrucache-backend` was written by `Josh Smeaton <josh.smeaton@gmail.com>`_.\n",
    'author': 'Josh Smeaton',
    'author_email': 'josh.smeaton@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kogan/django-lrucache-backend',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.4',
}


setup(**setup_kwargs)
